#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <bfd.h>
#include <sys/ptrace.h>
#include <signal.h>
#include <sys/user.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>      // for O_RDONLY

typedef unsigned char byte;
typedef unsigned int address;
typedef unsigned int word;

typedef struct {
	unsigned int value;
	char name[59];
	byte op;
} sym_t;

typedef struct {
	long low;
	long high;
} range_t;

static sym_t *active_syms;
static range_t range;

static int num_syms;
static int verbose = 1;
static char buf[256];

extern void sendGui(const char *str);
extern int service_gui(void);

#define printg(a)	sprintf a; \
					sendGui(buf)

#define eptrace(action,pid,addr,val)				\
	if ( (ptrace (action, pid, addr, val)) == -1) {	\
		perror ("ptrace");							\
		exit (1);									\
	}

#define IOP 0xf1	// The Illegal OP
#define MAX_SIGS 10	// Max number of authorized continued signals
					//	(prevents infinite loop)
					//	Set it to 0 to disable checking

/******************************************************************************/
/* BP_set sets a breakpoint by replacing the normal instruction
 * with IOP (0xf1), an illegal instruction which generate a SIGTRAP.
 *
 * Return value: the normal instruction, to be restored later
 */
inline byte BP_set (int pid, address offset)
{
	word prev=0;		// Previous value
	
	prev = ptrace (PTRACE_PEEKTEXT, pid, offset, 0);
	if (prev == -1) {	// Hoping that we won't encounter 0xffffffff at offset
		perror ("ptrace");
		printg ((buf, "Error in BP_set, on pid=%d and offset=0x%x.\n",
					pid, offset));
		exit (1);
	}

	eptrace (PTRACE_POKETEXT, pid, offset, (prev & 0xffffff00) | IOP);
	
	return (0xff & prev);	// Returns only the opcode byte
}

/* BP_unset unsets a BP previously set with BT_set.
 * It simply puts prev at offset to replace the 0xf1
 */
inline void BP_unset (int pid, address offset, byte prev)
{
	register word new;		// The new value we'll write

	new = ptrace (PTRACE_PEEKTEXT, pid, offset, 0);

	// Current value except for last byte
	eptrace (PTRACE_POKETEXT, pid, offset, (new & 0xffffff00) | prev);
}
/*****************************************************************************/

/*
 * Determines the code-range (low and upper border) and its size
 */
void CodeRange(bfd *abfd)
{
	asection* secp;
	long size;
	long *low = &(range.low);
	long *high = &(range.high);

	for (secp = abfd->sections; secp != NULL; secp = secp->next) {
/*		if (verbose) {
            printg((buf, "section: %s\tvma: %08llX\tsize: %08llX\tflags: %d\n",
			       secp->name,secp->vma,secp->size,secp->flags));
		}*/
		if (secp->flags & SEC_CODE) {
			if (!strcmp(secp->name,".text")) {
				*low  = secp->vma;
				*high = secp->vma+secp->size-1;
				break;
			}
		}
	}
	size = (*high - *low)+1;
	printg((buf,"Code from %08lX to %08lX (%ld bytes)\n\n",*low,*high,size));
}

void traceDetach(int pid)
{
	int i;
	sym_t *symPtr = active_syms;
	for(i = 0; i < num_syms; i++)
	{
		if (symPtr->op != IOP)
		{
			// Removes the BP
			BP_unset (pid, symPtr->value, symPtr->op);
		}
		symPtr++;
	}
    printg((buf,"Detaching from pid %d\n", pid));
	ptrace(PTRACE_DETACH, pid, 0, 0);
}

char *traceAttach(const char *fname, int pid)
{
	bfd *abfd;
    int status = 0;
    struct user_regs_struct uregs;
	int storage_needed, i, num_symbols, count;
	asymbol **symbol_table;
	char *symStr;
	char *charPtr;
	sym_t *symPtr;
	byte prev;

	bfd_init(); // magic
    abfd = bfd_openr(fname, NULL);
    assert(abfd != NULL);
    bfd_check_format(abfd, bfd_object);
	CodeRange(abfd);

	if (ptrace(PTRACE_ATTACH, pid, 0, 0) < 0)
		goto error;
    wait(&status);
	//if (WIFSTOPPED(status) && WSTOPSIG(status))
	//	printf("Yess really stopped");
	//ptrace(PTRACE_GETSIGINFO, pid, 0, &sinfo);
	//printf("si_signo is %i\n", sinfo.si_signo);
	//ptrace(PTRACE_ATTACH, pid, 0, 0);
    eptrace(PTRACE_GETREGS, pid, 0, &uregs);
    printg((buf, "Stopped at eip: 0x%x\n", (unsigned int)uregs.eip));
    //printf("Stopped at eip: 0x%x\n", (unsigned int)uregs.eip);


    storage_needed = bfd_get_symtab_upper_bound(abfd);
    assert(storage_needed >= 0);
	//printf("storage = %d\n", storage_needed);

    symbol_table = (asymbol**)malloc(storage_needed);
    assert(symbol_table != 0);
    num_symbols = bfd_canonicalize_symtab(abfd, symbol_table);
	printg((buf, "Total number of symbols in file = %d\n", num_symbols));
	if(num_symbols <= 0)
		goto error;
	active_syms = (sym_t*)malloc(num_symbols * sizeof(sym_t));
	if (active_syms == NULL)
		goto error;
	/*  Initialise op to an impossible opcode */
	symPtr = active_syms;
	for(i = 0; i < num_syms; i++)
	{
		symPtr->op = IOP;
		symPtr++;
	}

	symStr = (char *)malloc(num_symbols * 59);
	if (symStr == NULL)
		goto error;
	printg((buf, "Setting up breakpoints.. "));
	count = 0;
	symPtr = active_syms;
	charPtr = symStr;
    for(i = 0; i < num_symbols; i++)
	{
		asymbol *sym = symbol_table[i];
		if ((sym->flags & BSF_GLOBAL) && (sym->flags & BSF_FUNCTION))
		{
			size_t len;

			/* WARNING bfd_asymbol_value is a long long */
			symPtr->value = bfd_asymbol_value(sym);
			assert(strlen(bfd_asymbol_name(sym)) <= 55);
			strcpy ( symPtr->name, bfd_asymbol_name(sym) );
			len = strlen(symPtr->name);
			sprintf(charPtr, "%s ", symPtr->name);
			charPtr += len+1;
        	//printf("%s: %x\n", symPtr->name, symPtr->value);
			prev = BP_set (pid, symPtr->value);
			if (prev != IOP)
				symPtr->op = prev;
			else
			{
				printg ((buf, "\nWarning: tried to set same breakpoint (0x%x)"
								"twice.\n", (unsigned int) symPtr->value));
			}
			symPtr++; count++;
			//printf("symPtr is %p\n", symPtr);
		}
	}
	num_syms = count;
	printg((buf, "done\n%d active symbols/breakpoints\n", num_syms));
	free(symbol_table);
	bfd_close(abfd);
	return symStr;
error:
	free(symbol_table);
	bfd_close(abfd);
	printg((buf,"Error: detaching from process %d..\n", pid));
	ptrace(PTRACE_DETACH, pid, 0, 0);
	return NULL;
}

void traceFile(int pid, FILE *OUT)
{
    int i, status = 0; 
    struct user_regs_struct uregs;
	byte sig_count = -1;

/*  while(1){
	ptrace(PTRACE_SINGLESTEP, pid, 0, 0);
	wait(&status);
    eptrace(PTRACE_GETREGS, pid, 0, &uregs);
    printf("Stopped at eip: 0x%x\n", (unsigned int)uregs.eip);
	wait(&status);
	getchar();
  }*/
	printg((buf,"Starting pid %d..\n", pid));
	eptrace (PTRACE_CONT, pid, 0, 0);
	while(1)
	{
		if (service_gui())
		{
			kill (pid, SIGSTOP);
			wait(&status);
	    	printg((buf,"Pid %d stopped\n\n", pid));
			return;
		}
		if (waitpid (pid, &status, WNOHANG) > 0)
			break;
	}
	//test = ptrace (PTRACE_CONT, pid, 0, 0);
   	//printf("test = %d\n", test);
	//waitpid(pid, &status, WCONTINUED);

	while (!WIFEXITED(status)) {
		sym_t *symPtr;
		if (WIFSTOPPED(status)) {
			eptrace(PTRACE_GETREGS, pid, 0, &uregs);
			// Decrements eip to come back on the BPed instruction
			--uregs.eip;
			symPtr = active_syms;
			for(i = 0; i < num_syms; i++)
			{
				if(uregs.eip == symPtr->value)
					break;
				symPtr++;
			}
       		//printf("Stopped at %s: %x\n", symPtr->name, symPtr->value);

			if ((uregs.eip >= range.low) && (uregs.eip <= range.high) &&
						(symPtr->op != IOP)) {
				// Here it should be one of our breakpoints
				eptrace (PTRACE_SETREGS, pid, 0, &uregs);	// Decrements EIP
				// Removes the BP
				BP_unset (pid, uregs.eip, symPtr->op);
				//ipBuf[uregs.eip-codeLow]++;	// Keeps trace
				fprintf(OUT, "%s()\n", symPtr->name);
				ptrace(PTRACE_SINGLESTEP, pid, 0, 0);
				wait(&status);
				// And replace once more
				symPtr->op = BP_set (pid, symPtr->value);

				eptrace (PTRACE_CONT, pid, 0, 0);	// And continue.

			} else {	// Unknown one
				int sig = WSTOPSIG (status);
				++sig_count;
				if (verbose) 
				{
					printg ((buf,"Program stopped inside our code"
						"range, but not on one of our breakpoints "
						"(eip: 0x%x, opcode: %x, op: 0x%x, signal %i).\n",
						(unsigned int) uregs.eip+1,
						(unsigned int) ((ptrace (PTRACE_PEEKTEXT, pid, uregs.eip+1,
							0) & 0xff)),	// Not really uregs.eip-1
						(unsigned int) symPtr->op,
						sig));
				}
				// Infinite loop checking
				if (sig != SIGABRT && sig != SIGFPE && sig != SIGSEGV &&
					sig != SIGILL && sig != SIGTERM 
					&& (!MAX_SIGS || sig_count < MAX_SIGS)) {
					eptrace (PTRACE_CONT, pid, 0, 0);
				} else {
					if (verbose && (MAX_SIGS && sig_count >= MAX_SIGS))
					{
						printg ((buf, "Killing child due to too many "
									"signals (%i).\n", MAX_SIGS));
					}
					else if (verbose)
					{
						printg ((buf,"Child exiting on signal %d...\n", sig));
					}
					eptrace (PTRACE_KILL, pid, 0, 0);
					break;
				}
			}
		}
		if(service_gui())
		{
			wait(&status);
			if (WIFSTOPPED(status))
			{
				eptrace(PTRACE_GETREGS, pid, 0, &uregs);
				// Decrements eip to come back on the BPed instruction
				--uregs.eip;
				symPtr = active_syms;
				for(i = 0; i < num_syms; i++)
				{
					if(uregs.eip == symPtr->value)
						break;
					symPtr++;
				}
				eptrace (PTRACE_SETREGS, pid, 0, &uregs);	// Decrements EIP
				// Remove the BP
				BP_unset (pid, uregs.eip, symPtr->op);
				ptrace(PTRACE_SINGLESTEP, pid, 0, 0);
				wait(&status);
				// And replace once more
				symPtr->op = BP_set (pid, symPtr->value);
			}
	    	printg((buf,"Pid %d stopped\n\n", pid));
			return;
		}
		wait(&status);
	}
}

